#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import datetime
import os
import sys

import psycopg2
from lxml import etree
from pycapo import CapoConfig

profile = os.environ["CAPO_PROFILE"]


def getElement(root, element):
    return root.find(".//{http://www.w3schools.com/VLAarchive}" + element)


def getElementList(root, element):
    return root.findall(".//{http://www.w3schools.com/VLAarchive}" + element)


def connect(profile):
    """Connect to the PostgreSQL database server"""
    conn = None
    try:

        config = CapoConfig(profile=profile)
        user = config["metadataDatabase.jdbcUsername"]
        password = config["metadataDatabase.jdbcPassword"]
        url = config["metadataDatabase.jdbcUrl"]
        archiveurl = url.replace("jdbc:", "").replace("://", "://" + user + ":" + password + "@")

        # connect to the PostgreSQL server
        print("Connecting to " + archiveurl)
        conn = psycopg2.connect(archiveurl)

        return conn

    except (Exception, psycopg2.DatabaseError) as error:
        print(error)


def parseFile(file_name, log_file):
    # set up the log file:
    log = open(log_file, "w")

    # open and parse the file:
    try:
        input_file = open(file_name, "r")
        tree = etree.parse(input_file)
    except:
        print("Could not parse file {}!".format(file_name))
        log.write("could not parse file {}!".format(file_name))
        return 1

    # connect to the DB:
    try:
        conn = connect(profile)
        cur = conn.cursor()
        cur.execute("SELECT version()")
        db_version = cur.fetchone()
    except:
        print("Could not connect to the database!")
        log.write("could not connect to the database!")
        return 1

    # collect the values we need from the xml:
    root = tree.getroot()

    project = {}
    project["project_code"] = getElement(root, "observing_program_Id").text
    project["configuration"] = getElement(root, "configuration").text
    project["starttime"] = getElement(root, "start_time").text
    project["endtime"] = getElement(root, "end_time").text
    project["prop_expiration"] = 0
    project["prop_duration"] = 365

    log.write(str(project))

    if project["project_code"] is None:
        print("No project code found: exiting.")
        log.write("no project code found.")
        return 1

    # create the project, if necessary
    cur.execute("SELECT starttime, endtime from projects where project_code=%s", [project["project_code"]])
    existing_project = cur.fetchone()
    if existing_project is not None:
        # update the project, if necessary
        project["starttime"] = existing_project[0]
        project["endtime"] = existing_project[1]
        log.write("Project {} exists.".format(project["project_code"]))
    else:
        # create the project
        log.write("Creating Project:")
        log.write(
            "project code: %s, legacy_id: %s, title: %s, abstract: %s, start_time: %s, end_time: %s"
            % (
                project["project_code"],
                project["project_code"],
                "null",
                "null",
                project["starttime"],
                project["endtime"],
            )
        )
        cur.execute(
            "INSERT INTO projects (project_code, legacy_id, starttime, endtime) VALUES (%s, %s, %s, %s);",
            [project["project_code"], project["project_code"], project["starttime"], project["endtime"]],
        )
    author = {}
    author["name"] = getElement(root, "observer_name").text
    name = author["name"].split(" ", 1)

    # create the author, if necessary
    cur.execute("SELECT firstname, lastname from authors where project_code=%s", [project["project_code"]])
    existing_authors = cur.fetchone()
    if existing_authors is not None:
        # authors already exist
        log.write("Project %s has author %s %s." % (project["project_code"], existing_authors[0], existing_authors[1]))
    else:
        # create the author
        log.write("Creating Author %s:" % (author["name"]))
        cur.execute(
            "INSERT INTO authors (project_code, username, firstname, lastname, is_pi) VALUES (%s, %s, %s, %s, %s);",
            [project["project_code"], "unknown", name[0], name[1], True],
        )

    eb = {}
    eb["configuration"] = getElement(root, "configuration").text
    band_list = getElementList(root, "band")
    bands = ""
    for e in band_list:
        if e.text is not None:
            bands = bands + e.text + " "

    eb["band_code"] = bands

    filesize = os.stat(file_name[:-4] + ".exp").st_size

    # assume for now that all entries in a file belong to the same new EB:
    cur.execute(
        "insert into filegroups (project_code, type, datasize) values (%s, %s, %s) returning filegroup_id;",
        (project["project_code"], "execution_block", filesize),
    )
    filegroup_id = cur.fetchone()[0]

    # there is one .exp file per .xml file:
    filename = file_name.split("/")[-1][:-4] + ".exp"
    cur.execute(
        "insert into files (filegroup, filename, filesize, format, type, ngas_location, ngas_cluster, ngas_id)"
        " values(%s, %s, %s, %s, %s, %s, %s, %s)",
        (filegroup_id, filename, filesize, "exp", "raw", "DSOC", "DSOC", filename),
    )

    cur.execute(
        "insert into science_products (science_product_locator, science_product_type, metadata_ingestion_date, metadata_ingestion_version, filegroup_id, external_name) "
        "values ('uid://VLA/execblock/'||gen_random_uuid(), %s, %s, %s, %s, %s) returning science_product_locator;",
        ("Execution Block", datetime.datetime.now(), 1, filegroup_id, filename),
    )
    spl = cur.fetchone()[0]

    log.write("creating science_products_projects entry:")
    cur.execute(
        "insert into science_products_projects (science_product_locator, project_code) values(%s, %s)",
        (spl, project["project_code"]),
    )

    log.write(
        "Creating EB: {}, {}, {}, {}, {}, {}, {}, {}".format(
            filegroup_id,
            0,
            "VLA",
            project["project_code"],
            "DO Not Calibrate",
            eb["band_code"],
            eb["configuration"],
            True,
        )
    )
    cur.execute(
        "insert into execution_blocks (science_product_locator, filegroup_id, calibration_level, telescope, project_code"
        ", calibration_status, band_code, configuration, ingestion_complete) "
        "values (%s, %s, 0, 'VLA', %s, 'Do Not Calibrate', %s, %s, True) returning execution_block_id;",
        (spl, filegroup_id, project["project_code"], eb["band_code"], eb["configuration"]),
    )
    eb_id = cur.fetchone()[0]

    scan = {}

    subscans = getElementList(root, "scan")
    log.write("found %s scans".format(len(subscans)))
    for e in subscans:
        # create one scan entry and a subscan entry for each record:
        cur.execute(
            "insert into scans (execution_block_id, filegroup_id, max_bandwidth, min_bandwidth, polarization_code, max_frequency, min_frequency) "
            "values (%s, %s, %s, %s, %s, %s, %s) returning scan_id;",
            (eb_id, filegroup_id, 0, 0, 0, 0, 0),
        )
        scan_id = cur.fetchone()[0]

        scan_max_freq = 0.0
        scan_min_freq = sys.float_info.max
        scan_max_bw = 0.0
        scan_min_bw = sys.float_info.max
        scan_pol_code = 0
        eb_start = sys.float_info.max
        eb_end = 0.0

        subscan = {}
        subscan["starttime"] = getElement(e, "start_time").text
        subscan["endtime"] = getElement(e, "end_time").text
        subscan["ra"] = getElement(e, "right_ascension_J2000").text
        subscan["dec"] = getElement(e, "declination_J2000").text
        subscan["sourcename"] = getElement(e, "source_name").text
        subscan["sourcetype"] = "NO IDEA"
        subscan["obstype"] = ""
        subscan["exposure_time"] = 0.0
        subscan["integration_time"] = getElement(e, "integration_time").text
        subscan["receiver_id"] = 0
        subscan["backend"] = "none"

        if subscan["sourcename"] is None or len(subscan["sourcename"]) <= 0:
            subscan["sourcename"] = "Unknown"

        # polarizations are annoying and fiddly.
        # create a configuration for the subscan, and create a data_description linked to the configuration for each polarization
        # in the XML
        log.write("creating configuration")
        cur.execute(
            "insert into configurations (configuration, execution_block_id) values (%s, %s) returning configuration_id",
            (0, eb_id),
        )
        config_id = cur.fetchone()

        spectral_windows = getElementList(e, "spectral_window")
        for sw in spectral_windows:
            window = {}
            window["frequency"] = getElement(sw, "center_frequency").text
            window["bandwidth"] = getElement(sw, "bandwidth").text
            window["channels"] = getElement(sw, "number_of_channels").text
            window["polarization"] = getElement(sw, "polarizations").text
            polarization_id = 0
            if "LL" in window["polarization"]:
                polarization_id += 16
            if "LR" in window["polarization"]:
                polarization_id += 32
            if "RL" in window["polarization"]:
                polarization_id += 64
            if "RR" in window["polarization"]:
                polarization_id += 128

            # set some scan aggregate values:
            scan_max_freq = max(scan_max_freq, float(window["frequency"]))
            scan_min_freq = min(scan_min_freq, float(window["frequency"]))
            scan_max_bw = max(scan_max_bw, float(window["bandwidth"]))
            scan_min_bw = min(scan_min_bw, float(window["bandwidth"]))
            scan_pol_code = scan_pol_code | polarization_id  # bitwise or

            log.write(str(window))
            cur.execute(
                "insert into data_descriptions (bandwidth, frequency, polarization_id, configuration_id) values (%s, %s, %s, %s)",
                (window["bandwidth"], window["frequency"], polarization_id, config_id),
            )
        # end for spectral windows

        log.write(str(subscan))
        cur.execute(
            "insert into subscans (scan_id, obstype, starttime, endtime, sourcename, sourcetype, ra, dec, exposure_time"
            ", integration_time, receiver_id, backend, configuration_id)"
            "values (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s);",
            (
                scan_id,
                subscan["obstype"],
                subscan["starttime"],
                subscan["endtime"],
                subscan["sourcename"],
                subscan["sourcetype"],
                subscan["ra"],
                subscan["dec"],
                subscan["exposure_time"],
                subscan["integration_time"],
                subscan["receiver_id"],
                subscan["backend"],
                config_id,
            ),
        )

        # set some project aggregate values:
        project["starttime"] = min(float(project["starttime"]), float(subscan["starttime"]))
        project["endtime"] = max(float(project["endtime"]), float(subscan["endtime"]))
        eb_start = min(eb_start, float(subscan["starttime"]))
        eb_end = max(eb_end, float(subscan["endtime"]))

        if scan_min_bw == sys.float_info.max:
            scan_min_bw = 0
        if scan_min_freq == sys.float_info.max:
            scan_min_freq = 0

        cur.execute(
            "update scans set max_bandwidth=%s, min_bandwidth=%s, max_frequency=%s, min_frequency=%s, polarization_code=%s where scan_id=%s",
            (scan_max_bw, scan_min_bw, scan_max_freq, scan_min_freq, scan_pol_code, scan_id),
        )

    # end for subscans
    if eb_start == sys.float_info.max:
        eb_start = 0

    # update the eb with aggregated values:
    cur.execute(
        "update execution_blocks set starttime=%s, endtime=%s where execution_block_id=%s", (eb_start, eb_end, eb_id)
    )

    # update the project with aggregate values:
    cur.execute(
        "update projects set starttime=%s, endtime=%s, proprietary_expiration=mjd_to_timestamp(%s), last_addition = mjd_to_timestamp(%s), proprietary_duration=%s where project_code = %s",
        (
            project["starttime"],
            project["endtime"],
            float(project["endtime"]) + 365.0,
            project["endtime"],
            365,
            project["project_code"],
        ),
    )

    cur.execute("commit;")
    return 0
